/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2006 Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Sam Lantinga
    slouken@libsdl.org
*/
#include "SDL_config.h"
#include "SDL_video.h"
#include "SDL_mouse.h"

#include "../SDL_sysvideo.h"
#include "../SDL_pixels_c.h"
#include "../../events/SDL_events_c.h"

#include "SDL_iphonevideo.h"
#include "SDL_iphoneevents.h"

#define IPHONEVID_DRIVER_NAME "iphone"

static int iPhone_VideoInit(_THIS, SDL_PixelFormat *vformat);
static SDL_Rect **iPhone_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags);
static SDL_Surface *iPhone_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags);
static int iPhone_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors);
static void iPhone_VideoQuit(_THIS);
static void iPhone_UpdateRects(_THIS, int numrects, SDL_Rect *rects);
static int iPhone_Available(void) { return 1; }
static int iPhone_LockHWSurface(_THIS, SDL_Surface *surface);
static void iPhone_UnlockHWSurface(_THIS, SDL_Surface *surface);

static void iPhone_DeleteDevice(SDL_VideoDevice *device)
{
	SDL_free(device->hidden);
	SDL_free(device);
}

static SDL_VideoDevice *iPhone_CreateDevice(int devindex)
{
	SDL_VideoDevice *device;

	device = (SDL_VideoDevice *)SDL_malloc(sizeof(SDL_VideoDevice));
	if (device) 
	{
		SDL_memset(device, 0, (sizeof *device));
		device->hidden = (struct SDL_PrivateVideoData *) SDL_malloc((sizeof *device->hidden));
	}
	
	if ( (device == NULL) || (device->hidden == NULL) ) 
	{
		SDL_OutOfMemory();
		if ( device ) SDL_free(device);
		return 0;
	}

	SDL_memset(device->hidden, 0, (sizeof *device->hidden));
	device->hidden->_view = [UIApp getView];

	device->VideoInit = iPhone_VideoInit;
	device->ListModes = iPhone_ListModes;
	device->SetVideoMode = iPhone_SetVideoMode;
	device->CreateYUVOverlay = NULL;
	device->SetColors = iPhone_SetColors;
	device->UpdateRects = iPhone_UpdateRects;
	device->VideoQuit = iPhone_VideoQuit;
	device->AllocHWSurface = NULL;
	device->CheckHWBlit = NULL;
	device->FillHWRect = NULL;
	device->SetHWColorKey = NULL;
	device->SetHWAlpha = NULL;
	device->LockHWSurface = iPhone_LockHWSurface;
	device->UnlockHWSurface = iPhone_UnlockHWSurface;
	device->FlipHWSurface = NULL;
	device->FreeHWSurface = NULL;
	device->SetCaption = NULL;
	device->SetIcon = NULL;
	device->IconifyWindow = NULL;
	device->GrabInput = NULL;
	device->GetWMInfo = NULL;
	device->InitOSKeymap = iPhone_InitOSKeymap;
	device->PumpEvents = iPhone_PumpEvents;
	device->free = iPhone_DeleteDevice;
	
	return device;
}

VideoBootStrap iPhone_bootstrap = 
{
	IPHONEVID_DRIVER_NAME, "iPhone video driver",
	iPhone_Available, iPhone_CreateDevice
};

int iPhone_VideoInit(_THIS, SDL_PixelFormat *vformat)
{
	vformat->BitsPerPixel = 16;
	vformat->Rmask = 0xF800;
	vformat->Gmask = 0x07E0;
	vformat->Bmask = 0x001F;

	return 0;
}

SDL_Rect** iPhone_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
	return (SDL_Rect**)-1;
/*
	static SDL_Rect rect;
	static SDL_Rect* rectArray[2] = {&rect, 0};
	
	CGRect cgRect = [UIHardware fullScreenApplicationContentRect];
	rect.x = 0;
	rect.y = 0;
	rect.w = cgRect.size.width;
	rect.h = cgRect.size.height / 2;
	
	return (SDL_Rect**)rectArray;
*/
}

SDL_Surface *iPhone_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags)
{
	if(bpp != 16)
		return NULL;

	NSDictionary* size = [[NSDictionary alloc] initWithObjectsAndKeys:[NSNumber numberWithInt:width], @"w", [NSNumber numberWithInt:height], @"h", nil];
	[this->hidden->_view performSelectorOnMainThread:@selector(setScreenSize:) withObject:size waitUntilDone:YES];
	[size release];
		
	current->flags = (SDL_FULLSCREEN|SDL_HWSURFACE);
	current->w = width;
	current->h = height;
	current->pitch = current->w * 2;
	current->pixels = CoreSurfaceBufferGetBaseAddress([this->hidden->_view getSurface]);
	SDL_ReallocFormat(current, bpp, 0xF800, 0x07E0, 0x001F, 0);
	
	return current;
}

int iPhone_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
{
	return(1);
}

void iPhone_VideoQuit(_THIS)
{
	this->screen->pixels = NULL;
}

static void noise_test(_THIS)
{
	CGRect frame = [this->hidden->_view frame];
	unsigned short* p = CoreSurfaceBufferGetBaseAddress([this->hidden->_view getSurface]);
	int i,j;
	for(i=0;i<frame.size.height;i++)
	{
		for(j=0;j<frame.size.width;j++,p++)
		{
			*p=(rand()%2) ?	 0xF800 : 0x07E0;;
		}
	}
}

static void iPhone_UpdateRects(_THIS, int numrects, SDL_Rect *rects)
{
//	noise_test(this);
	[this->hidden->_view performSelectorOnMainThread:@selector(setNeedsDisplay) withObject:nil waitUntilDone: NO];
}


static int iPhone_LockHWSurface(_THIS, SDL_Surface *surface)
{
	CoreSurfaceBufferLock([this->hidden->_view getSurface], 3);
}

static void iPhone_UnlockHWSurface(_THIS, SDL_Surface *surface)
{
    CoreSurfaceBufferUnlock([this->hidden->_view getSurface]);
}

/* SDL_iPhoneView */
@implementation SDL_iPhoneView

- (id)initWithFrame:(struct CGRect)frame 
{
	[super initWithFrame: frame];
	
	_events = nil;
	_lock = [[NSLock alloc] init];
	[_lock retain];

	_screenSurface = NULL;
	_screenLayer   = [[LKLayer layer] retain];
	[_screenLayer setFrame: CGRectMake(0.0f, 0.0f, frame.size.width, frame.size.height)];
	[_screenLayer setOpaque: YES];
	[[self _layer] addSublayer: _screenLayer];
	
	_frameRect = frame;
	_screenSize.width  = 0;
	_screenSize.height = 0;
	
	return self;
}

- (void)setScreenSize:(NSDictionary*)size
{
	int w = _screenSize.width  = [[size objectForKey:@"w"] intValue];
	int h = _screenSize.height = [[size objectForKey:@"h"] intValue];

	int pitch = w * 2, allocSize = 2 * w * h;
	char *pixelFormat = "565L";

	NSDictionary* dict = [[NSDictionary alloc] initWithObjectsAndKeys:
		kCFBooleanTrue, kCoreSurfaceBufferGlobal,
		 @"PurpleGFXMem", kCoreSurfaceBufferMemoryRegion,
		[NSNumber numberWithInt: pitch], kCoreSurfaceBufferPitch,
		[NSNumber numberWithInt: w], kCoreSurfaceBufferWidth,
		[NSNumber numberWithInt: h], kCoreSurfaceBufferHeight,
		[NSNumber numberWithInt: *(int*)pixelFormat], kCoreSurfaceBufferPixelFormat,
		[NSNumber numberWithInt: allocSize], kCoreSurfaceBufferAllocSize,
		nil
	];

	_screenSurface = CoreSurfaceBufferCreate((CFDictionaryRef)dict);
	CoreSurfaceBufferLock(_screenSurface, 3);
	[_screenLayer setContents: _screenSurface];
	CoreSurfaceBufferUnlock(_screenSurface);
	[dict release];
}

- (void)dealloc 
{
	[_events release];
	[_lock release];
	[_screenLayer release];
	[super dealloc];
}

- (void)lock
{
	[_lock lock];
}

- (void)unlock
{
	[_lock unlock];
}

- (NSArray*)getEvents
{
	[self lock];
	NSArray* events = _events;
	_events = nil;	
	[self unlock];
	
	[events autorelease];

	return events;
}

- (void)addEvent:(NSDictionary*)event
{
	[self lock];
	
	if(_events == nil)
		_events = [[NSMutableArray alloc] init];
		
	[_events addObject: event];
	
	[self unlock];
}

- (void)gestureChanged:(GSEvent*)event
{
	printf("gestureChanged!\n");
}

- (void)gestureEnded:(GSEvent*)event
{
	printf("gestureEnded!\n");
}

- (void)gestureStarted:(GSEvent*)event
{
	printf("gestureStarted!\n");
}

- (void)keyDown:(GSEvent*)event
{
	[self addEvent:
		[[NSDictionary alloc] initWithObjectsAndKeys:
		 @"keyDown", @"type",
		 nil
		]
	];
}

- (void)keyUp:(GSEvent*)event
{
	[self addEvent:
		[[NSDictionary alloc] initWithObjectsAndKeys:
		  @"keyUp", @"type",
		 nil
		]
	];
}

/*
- (void)mouseMoved:(GSEvent*)event
{
	struct CGRect rect = GSEventGetLocationInWindow(event);
	
	rect.origin.x -= _frameRect.origin.x;
	rect.origin.y -= _frameRect.origin.y;
	rect.origin.x  = rect.origin.x * _screenSize.width  / _frameRect.size.width;
	rect.origin.y  = rect.origin.y * _screenSize.height / _frameRect.size.height;

	//printf("x=%d y=%d\n", (int)rect.origin.x, (int)rect.origin.y);

	[self addEvent:
		[[NSDictionary alloc] initWithObjectsAndKeys:
		 @"mouseMoved", @"type",
		 [NSNumber numberWithInt:(int)rect.origin.x], @"x",
		 [NSNumber numberWithInt:(int)rect.origin.y], @"y",
		 nil
		]
	];
}

- (void)mouseDown:(GSEvent*)event
{
	struct CGRect rect = GSEventGetLocationInWindow(event);

	rect.origin.x -= _frameRect.origin.x;
	rect.origin.y -= _frameRect.origin.y;
	rect.origin.x  = rect.origin.x * _screenSize.width  / _frameRect.size.width;
	rect.origin.y  = rect.origin.y * _screenSize.height / _frameRect.size.height;

	//printf("x=%d y=%d\n", (int)rect.origin.x, (int)rect.origin.y);

	[self addEvent:
		[NSDictionary dictionaryWithObjectsAndKeys:
		 @"mouseDown", @"type",
		 [NSNumber numberWithInt:(int)rect.origin.x], @"x",
		 [NSNumber numberWithInt:(int)rect.origin.y], @"y",
		 nil
		]
	];
}
*/
- (void)mouseDown:(GSEvent*)event
{
	struct CGRect rect = GSEventGetLocationInWindow(event);
	[self addEvent:
		[[NSDictionary alloc] initWithObjectsAndKeys:
		 @"mouseDown", @"type",
		 [NSNumber numberWithInt:(int)rect.origin.x], @"x",
		 [NSNumber numberWithInt:(int)rect.origin.y], @"y",
		 nil
		]
	];
}

- (void)mouseDragged:(GSEvent*)event
{
	struct CGRect rect = GSEventGetLocationInWindow(event);
	[self addEvent:
		[[NSDictionary alloc] initWithObjectsAndKeys:
		 @"mouseDragged", @"type",
		 [NSNumber numberWithInt:(int)rect.origin.x], @"x",
		 [NSNumber numberWithInt:(int)rect.origin.y], @"y",
		 nil
		]
	];
}

/*
- (void)mouseUp:(GSEvent*)event
{
	struct CGRect rect = GSEventGetLocationInWindow(event);

	rect.origin.x -= _frameRect.origin.x;
	rect.origin.y -= _frameRect.origin.y;
	rect.origin.x  = rect.origin.x * _screenSize.width  / _frameRect.size.width;
	rect.origin.y  = rect.origin.y * _screenSize.height / _frameRect.size.height;

	//printf("x=%d y=%d\n", (int)rect.origin.x, (int)rect.origin.y);

	[self addEvent:
		[[NSDictionary alloc] initWithObjectsAndKeys:
		 @"mouseUp", @"type",
		 [NSNumber numberWithInt:(int)rect.origin.x], @"x",
		 [NSNumber numberWithInt:(int)rect.origin.y], @"y",
		 nil
		]
	];
}
*/

- (void)mouseUp:(GSEvent*)event
{
	struct CGRect rect = GSEventGetLocationInWindow(event);
	[self addEvent:
		[[NSDictionary alloc] initWithObjectsAndKeys:
		 @"mouseUp", @"type",
		 [NSNumber numberWithInt:(int)rect.origin.x], @"x",
		 [NSNumber numberWithInt:(int)rect.origin.y], @"y",
		 nil
		]
	];}
/*
- (void)mouseDragged:(GSEvent*)event
{
	struct CGRect rect = GSEventGetLocationInWindow(event);
	
	rect.origin.x -= _frameRect.origin.x;
	rect.origin.y -= _frameRect.origin.y;
	rect.origin.x  = rect.origin.x * _screenSize.width  / _frameRect.size.width;
	rect.origin.y  = rect.origin.y * _screenSize.height / _frameRect.size.height;

	//printf("x=%d y=%d\n", (int)rect.origin.x, (int)rect.origin.y);

	[self addEvent:
		[NSDictionary dictionaryWithObjectsAndKeys:
		 @"mouseDragged", @"type",
		 [NSNumber numberWithInt:(int)rect.origin.x], @"x",
		 [NSNumber numberWithInt:(int)rect.origin.y], @"y",
		 nil
		]
	];
}
*/
- (void)mouseEntered:(GSEvent*)event
{
	[self addEvent:
		[[NSDictionary alloc] initWithObjectsAndKeys:
		 @"mouseEntered", @"type",
		 nil
		]
	];
}

- (void)mouseExited:(GSEvent*)event
{
	[self addEvent:
		[[NSDictionary alloc] initWithObjectsAndKeys:
		 @"mouseExited", @"type",
		 nil
		]
	];
}

- (CoreSurfaceBufferRef)getSurface
{
	return _screenSurface;
}

- (void)drawRect:(CGRect)frame
{
}

@end

